@@ -77,8 +77,8 @@ def cloud_tpu_init() -> None:
   running_in_cloud_tpu_vm = True
 
   os.environ.setdefault('GRPC_VERBOSITY', 'ERROR')
-  os.environ['TPU_ML_PLATFORM'] = 'JAX'
-  os.environ['TPU_ML_PLATFORM_VERSION'] = version.__version__
+  os.environ.setdefault('TPU_ML_PLATFORM', 'JAX')
+  os.environ.setdefault('TPU_ML_PLATFORM_VERSION', version.__version__)
   os.environ.setdefault('ENABLE_RUNTIME_UPTIME_TELEMETRY', '1')
   os.environ["LIBTPU_INIT_ARGS"] = os.environ.get("LIBTPU_INIT_ARGS","") + " --xla_tpu_use_enhanced_launch_barrier=true"
 
