@@ -23,14 +23,11 @@ from jax._src import dispatch
 from jax._src import config
 from jax._src import dtypes
 from jax._src.util import safe_zip
-from jax._src.lib import xla_client
 
 zip, unsafe_zip = safe_zip, zip
 
 import numpy as np
 
-xops = xla_client.ops
-
 def _input_dtype(x, *_, **__):
   return dtypes.canonicalize_dtype(x.dtype, allow_extended_dtype=True)
 
@@ -59,11 +56,11 @@ def standard_abstract_eval(prim, shape_rule, dtype_rule, weak_type_rule,
     out = prim.impl(*[x.val for x in avals], **kwargs)
     return core.ConcreteArray(out.dtype, out, weak_type=weak_type)
   elif least_specialized is core.ShapedArray:
-    out_sharding = (sharding_rule(*avals, **kwargs)
-                    if config.sharding_in_types.value else None)
-    return core.ShapedArray(shape_rule(*avals, **kwargs),
-                            dtype_rule(*avals, **kwargs), weak_type=weak_type,
-                            sharding=out_sharding)
+    return core.ShapedArray(
+        shape_rule(*avals, **kwargs), dtype_rule(*avals, **kwargs),
+        weak_type=weak_type,
+        sharding=(sharding_rule(*avals, **kwargs)
+                  if config.sharding_in_types.value else None))
   elif least_specialized is core.DShapedArray:
     shape = shape_rule(*avals, **kwargs)
     ty = (core.ShapedArray if all(type(d) is int for d in shape)
@@ -96,13 +93,6 @@ def standard_multi_result_abstract_eval(
   else:
     raise TypeError(avals, least_specialized)
 
-def standard_translate(prim):
-  xla_opname = ''.join(term.capitalize() for term in prim.name.split('_'))
-  op = getattr(xops, xla_opname)
-  def translation_rule(ctx, avals_in, avals_out, *args, **kwargs):
-    del ctx, avals_in, avals_out
-    return [op(*args, **kwargs)]
-  return translation_rule
 
 def _standard_weak_type_rule(*avals, **kwargs):
   return all(aval.weak_type for aval in avals)
