@@ -12,7 +12,8 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-# flake8: noqa: F401
+# ruff: noqa: F401
 
 import jaxlib.mlir.ir as ir
 import jaxlib.mlir.passmanager as passmanager
+from jaxlib.mlir._mlir_libs import register_jax_dialects
