@@ -19,7 +19,7 @@ from collections.abc import Callable, Iterable, Sequence
 import dataclasses
 from functools import partial, reduce
 import itertools
-from typing import Any
+from typing import Any, Literal
 
 import jax
 from jax import lax
@@ -254,7 +254,7 @@ def _pallas_call_impl_interpret(
   num_inout_blocks = len(block_args) + len(out)
   grid_start_indices = (jnp.int32(0),) * len(grid)
   if grid:
-    num_iterations = reduce(jnp.multiply, grid)
+    num_iterations = reduce(jnp.multiply, grid)  # type: ignore[arg-type]
   else:
     # Base case is always one iteration when grid is ()
     num_iterations = 1
@@ -279,6 +279,10 @@ def _pallas_call_impl_interpret(
 
     carry_consts_ins, scratch = split_list(carry_blocks, [num_inout_blocks])
     with pallas_core.grid_env(local_grid_env):
+      for s in scalars:
+        if isinstance(s.dtype, jax_core.bint):
+          aval = jax_core.get_aval(s)
+          s.aval = aval.update(dtype=jnp.int32)
       start_indices = [
           None if bm is None else bm.compute_start_indices_interpret(loop_idx, *scalars)
           for bm in grid_mapping.block_mappings]
@@ -293,10 +297,6 @@ def _pallas_call_impl_interpret(
           len(blocks),
           len(scratch_values),
       )
-      for s in scalars:
-        aval = jax_core.get_aval(s)
-        if isinstance(aval, jax_core.DShapedArray):
-          s.aval = aval.update(dtype=jnp.int32)
 
       blocks = jax_core.eval_jaxpr(
           discharged_jaxpr, discharged_consts, *scalars, *blocks, *scratch
@@ -362,6 +362,7 @@ def _pallas_call_jvp_rule(
     compiler_params: Any,
     cost_estimate: CostEstimate | None,
     out_avals: tuple[jax_core.AbstractValue, ...],
+    backend: _Backend | None,
 ):
   if grid_mapping.num_dynamic_grid_bounds:
     raise NotImplementedError("interpret with dynamic grid bounds unsupported")
@@ -425,7 +426,8 @@ def _pallas_call_jvp_rule(
       input_output_aliases=(),
       compiler_params=compiler_params,
       cost_estimate=jvp_cost_estimate,
-      out_avals=(*out_avals, *out_avals)
+      out_avals=(*out_avals, *out_avals),
+      backend=backend,
   )
   out_primals, out_tangents = split_list(out_flat, [len(out_flat) // 2])
   return out_primals, out_tangents
@@ -437,10 +439,11 @@ ad.primitive_jvps[pallas_call_p] = _pallas_call_jvp_rule
 def _batch_block_mapping(
     grid_mapping: GridMapping,
     axis_size: int,
+    for_ragged: bool,
     aval: jax_core.ShapedArray,
     dim: int | batching.NotMapped,
     block_mapping: BlockMapping,
-    for_ragged: bool,
+    ragged_axis_values,
 ) -> BlockMapping:
   def _block_map_function(new_idx, *args):
     if for_ragged:
@@ -466,14 +469,7 @@ def _batch_block_mapping(
   if for_ragged:
     if isinstance(dim, batching.RaggedAxis):
       assert for_ragged, "Ragged axis not supported for non-ragged batching."
-      _, _, ragged_axis_length = _ragged_axis_parts(dim)
-      aval = jax_core.get_aval(ragged_axis_length).update(dtype=jnp.int32)
-      if isinstance(aval, jax_core.DShapedArray):
-        aval = jax_core.ShapedArray(aval.shape, aval.dtype, aval.weak_type)
-      lengths_aval = pallas_core.AbstractMemoryRef(
-          aval,
-          pallas_core.MemorySpace.INDEX,
-      )
+      _, _, _, lengths_aval = ragged_axis_values
       idx_avals = [*idx_avals, lengths_aval]
     else:
       i32_aval_memref = pallas_core.AbstractMemoryRef(
@@ -538,6 +534,12 @@ def _broadcast_input_output_aliases(
   for input_index, _ in input_output_aliases:
     dim = dims_[input_index]
     dims_[input_index] = 0
+    if isinstance(dim, batching.RaggedAxis):
+      stacked_axis = dim.stacked_axis
+      if stacked_axis != 0:
+        raise NotImplementedError("Ragged aliasing on non 0 dim NYI")
+      return tuple(args_), tuple(dims_)
+
     if dim is batching.not_mapped:
       args_[input_index] = batching.broadcast(args_[input_index], axis_size, 0)
     elif dim != 0:
@@ -560,6 +562,7 @@ def _batch_with_explicit_loop(
     compiler_params: Any,
     cost_estimate: CostEstimate | None,
     out_avals: tuple[jax_core.AbstractValue, ...],
+    backend: _Backend | None,
 ):
   """Batch the pallas_call by calling it in loop over the batch size.
 
@@ -627,6 +630,7 @@ def _batch_with_explicit_loop(
         compiler_params=compiler_params,
         cost_estimate=cost_estimate,
         out_avals=out_avals,
+        backend=backend,
     )
     for i, batch_out_array in enumerate(batch_out):
       state[i] = jax.lax.dynamic_update_index_in_dim(
@@ -643,16 +647,6 @@ def _batch_with_explicit_loop(
   return result, (0,) * len(result)
 
 
-def _ragged_axis_parts(dim: batching.RaggedAxis) -> tuple[int, int, int]:
-  stacked_axis = dim.stacked_axis
-  ragged_axes = dim.ragged_axes
-  if len(ragged_axes) != 1:
-    raise ValueError("Multiple ragged axes not yet implemented.")
-  ragged_axis_dim = ragged_axes[0][0]
-  ragged_axis_length = ragged_axes[0][1]
-  return stacked_axis, ragged_axis_dim, ragged_axis_length
-
-
 def _pallas_call_batching_rule(
     args,
     dims,
@@ -666,6 +660,7 @@ def _pallas_call_batching_rule(
     compiler_params: Any,
     cost_estimate: CostEstimate | None,
     out_avals: tuple[jax_core.AbstractValue, ...],
+    backend: _Backend | None,
 ):
   def _maybe_squeeze_out_bdim(
       x: jax.Array, bdim: int | batching.NotMapped
@@ -674,21 +669,10 @@ def _pallas_call_batching_rule(
       return x
     return jnp.squeeze(x, axis=bdim)
 
-  all_ragged_axes = [d for d in dims if isinstance(d, batching.RaggedAxis)]
-  if len(all_ragged_axes) > 1:
-    raise ValueError("Multiple ragged dimensions not yet implemented.")
-
-  if all_ragged_axes:
-    stacked_axis, ragged_axis_dim, ragged_axis_length = _ragged_axis_parts(
-        all_ragged_axes[0]
-    )
-  else:
-    stacked_axis, ragged_axis_dim, ragged_axis_length = None, None, None
-
   def get_size(i, x, d):
     if not isinstance(d, batching.RaggedAxis):
       return x.shape[d]
-    return x.aval.shape[i]
+    return x.aval.shape[d.stacked_axis]
 
   (axis_size,) = {
       get_size(i=i, x=x, d=d)
@@ -709,6 +693,7 @@ def _pallas_call_batching_rule(
         compiler_params=compiler_params,
         cost_estimate=cost_estimate,
         out_avals=out_avals,
+        backend=backend,
     )
     return [jnp.expand_dims(x, 0) for x in out], (0,) * len(out)
 
@@ -742,6 +727,7 @@ def _pallas_call_batching_rule(
         compiler_params=compiler_params,
         cost_estimate=cost_estimate,
         out_avals=out_avals,
+        backend=backend,
     )
   else:
     pass  # No dynamic grid dimensions
@@ -776,6 +762,7 @@ def _pallas_call_batching_rule(
           compiler_params=compiler_params,
           cost_estimate=cost_estimate,
           out_avals=out_avals,
+          backend=backend,
       )
 
   if not dims:
@@ -793,35 +780,49 @@ def _pallas_call_batching_rule(
       args, dims, input_output_aliases=input_output_aliases, axis_size=axis_size
   )
 
+  # Each dim either has data about its ragged axis, or None
+  ragged_axis_values = []
+  for d in dims:
+    if isinstance(d, batching.RaggedAxis):
+      stacked_axis, ragged_axis_dim, ragged_axis_length = (
+          batching._ragged_axis_parts(d)
+      )
+      aval = jax_core.get_aval(ragged_axis_length).update(dtype=jnp.int32)
+      if isinstance(aval, jax_core.DShapedArray):
+        aval = jax_core.ShapedArray(aval.shape, aval.dtype, aval.weak_type)
+      lengths_aval = pallas_core.AbstractMemoryRef(
+          aval,
+          pallas_core.MemorySpace.INDEX,
+      )
+      # TODO(mvoz): Give this its own type
+      ragged_axis_values.append(
+          (stacked_axis, ragged_axis_dim, ragged_axis_length, lengths_aval)
+      )
+    else:
+      ragged_axis_values.append(None)  # type: ignore[arg-type]
+
   all_dims = list(dims) + [0] * grid_mapping.num_outputs
+  ragged_axis_values = ragged_axis_values + [None] * grid_mapping.num_outputs  # type: ignore[list-item]
 
   num_index_operands = grid_mapping.num_index_operands
   num_scratch_operands = grid_mapping.num_scratch_operands
 
-  lengths_aval = None
-  if ragged_axis_length is not None:
-    aval = jax_core.get_aval(ragged_axis_length).update(dtype=jnp.int32)
-    if isinstance(aval, jax_core.DShapedArray):
-      aval = jax_core.ShapedArray(aval.shape, aval.dtype, aval.weak_type)
-    lengths_aval = pallas_core.AbstractMemoryRef(
-        aval,
-        pallas_core.MemorySpace.INDEX,
-    )
-
   # Only add a batch dimension for the avals that actually have a grid mapping.
   # This excludes scalar prefetch inputs (the first in the list) and scratch
   # operands (the last in the list).
   avals_to_batch = avals[num_index_operands:(len(avals) - num_scratch_operands)]
+
   batched_block_mappings = map(
       partial(
           _batch_block_mapping,
           grid_mapping,
           axis_size,
-          for_ragged=lengths_aval is not None,
+          any(ragged_axis_values),
       ),
       avals_to_batch,
       all_dims[num_index_operands:],
       block_mappings,
+      ragged_axis_values[num_index_operands:],
   )
 
   index_map_tree_args, index_map_tree_kwargs = grid_mapping.index_map_tree.unflatten(
@@ -829,6 +830,17 @@ def _pallas_call_batching_rule(
   assert not index_map_tree_kwargs
   batched_index_map_args = (pallas_core.index_map_grid_aval,) + index_map_tree_args
 
+  lengths_aval = None  # type: ignore[assignment]
+
+  # Check all the ragged axis values, ensure their raggedness pattern
+  # is identical (consider moving this check up!)
+  for rav in ragged_axis_values:
+    if rav is not None:
+      if lengths_aval is None:
+        lengths_aval = rav[3]
+      else:
+        assert lengths_aval == rav[3], "NYI - different lengths in ragged batch"
+
   if lengths_aval:
     batched_index_map_args = batched_index_map_args + (lengths_aval,)
     num_index_operands += 1
@@ -854,6 +866,14 @@ def _pallas_call_batching_rule(
   else:
     batched_cost_estimate = None
 
+  # Start the ragged handling code
+  # Here, we:
+  #  - Rewrite the indexer to save memory (skip indices outside the ragged bounds)
+  #  - Rewrite the kernel to save compute (skip elements outside the ragged bounds)
+  #  - Update various internal structures/metadata to account for the new
+  #    block spec.
+  #  - Set the hacky flag of ragged_originating on the mapping, to signal to
+  #    the lowering code to treat mapped dimensions as part of the user grid.
   if lengths_aval:
     batched_grid_mapping = batched_grid_mapping.replace(
         get_grid_indices=lambda indices, maybe_include_mapped_dims: indices,
@@ -868,53 +888,102 @@ def _pallas_call_batching_rule(
     # we write 0s. This is useful for debugging, as certain lowering paths
     # like mosaic will write the last data as passthrough, leading to
     # potentially confusing results.
-    debug_zero_fill_counterfactual = debug
-
-    first_block_mapping = batched_grid_mapping.block_mappings[0]
+    block_mapped_dim_idxs = []
     for block_mapping in batched_grid_mapping.block_mappings:
-      # This invariant may already be checked elsewhere, but lets reaffirm it
-      assert block_mapping.block_shape == first_block_mapping.block_shape, (
-          f"block_mapping.block_shape: {block_mapping.block_shape}, "
-          f"first_block_mapping.block_shape: {first_block_mapping.block_shape}"
-      )
-      assert (
-          block_mapping.array_shape_dtype
-          == first_block_mapping.array_shape_dtype
-      ), (
-          f"block_mapping.array_shape_dtype: {block_mapping.array_shape_dtype},"
-          " first_block_mapping.array_shape_dtype:"
-          f" {first_block_mapping.array_shape_dtype}"
-      )
+      mapped_dim_idxs = []
+      for i, d in enumerate(block_mapping.block_shape):
+        if d is pallas_core.mapped:
+          mapped_dim_idxs.append(i)
+        else:
+          mapped_dim_idxs.append(None)  # type: ignore[arg-type]
+      block_mapped_dim_idxs.append(mapped_dim_idxs)
 
-    mapped_dim_idxs = [
-        i
-        for i, d in enumerate(first_block_mapping.block_shape)
-        if d is pallas_core.mapped
-    ]
-    assert len(mapped_dim_idxs) == 1
-    mapped_dim_idx = mapped_dim_idxs[0]
-    if stacked_axis != mapped_dim_idx:
-      raise ValueError(
-          f"Expected mapped dim to be {stacked_axis}, but got {mapped_dim_idx}"
-      )
+    mapped_dim_idx = None
+    for rav, mapped_dim_idxs in zip(ragged_axis_values, block_mapped_dim_idxs):
+      if rav is not None:
+        stacked_axis = rav[0]
+        if mapped_dim_idx is None:
+          mapped_dim_idx = mapped_dim_idxs[stacked_axis]
+          if mapped_dim_idxs[stacked_axis] is None:
+            raise ValueError(
+                f"Expected mapped dim to be {stacked_axis}, but got"
+                f" {mapped_dim_idxs[stacked_axis]}"
+            )
+        else:
+          assert mapped_dim_idx == mapped_dim_idxs[stacked_axis], (
+              f"Different mapped dims - expected {mapped_dim_idx}, but got"
+              f" {mapped_dim_idxs[stacked_axis]}"
+          )
 
-    assert ragged_axis_dim is not None, "Invariant violation"
     # This is the blockspec size of the dimension
-    val_at_ragged_dim = first_block_mapping.block_shape[ragged_axis_dim]
+    block_shapes = [b.block_shape for b in batched_grid_mapping.block_mappings]
+
+    # Parse out the operations from the jaxpr to determine how to mask the output
+    # NOTE! while this *could* be a default dict of None, and None is sound, as
+    # it denotes that there is no raggedness for the given var, we explicitly
+    # do not do this, so as to get better signal on implementation of rules
+    # A misimplemented rule that does not account for new vars being introduced
+    # will result in an error on the next op using the new var. The benefit of
+    # of forcing implementers to account for all outputs and intermediaries is
+    # a very nice one.
+
+    var_to_raggedness = {}
+    for invar, rav in zip(jaxpr.invars, ragged_axis_values):
+      var_to_raggedness[invar] = rav
+
+    for eqn in jaxpr.eqns:
+      prim = eqn.primitive
+      if prim not in batching.ragged_prop_rules:
+        raise NotImplementedError(f"Not implemented - ragged prop for {prim}")
+      rule = batching.ragged_prop_rules[prim]
+
+      invar_raggedness = [
+          (
+              var_to_raggedness.get(invar, None)
+              if isinstance(invar, jax_core.Var)
+              else None
+          )
+          for invar in eqn.invars
+      ]
+      invar_raggedness, outvar_raggedness = rule(invar_raggedness, eqn.outvars)
+
+      for invar, rav in zip(eqn.invars, invar_raggedness):  # type: ignore[assignment]
+        if isinstance(invar, jax_core.Var):
+          var_to_raggedness[invar] = rav
+      for outvar, rav in zip(eqn.outvars, outvar_raggedness):
+        if isinstance(outvar, jax_core.Var):
+          var_to_raggedness[outvar] = rav
+
+    for pos, invar in enumerate(jaxpr.invars):
+      ragged_axis_values[pos] = var_to_raggedness[invar]
+
+    per_input_ragged_axis_dim = []
+    for rav in ragged_axis_values:
+      if rav is not None:
+        per_input_ragged_axis_dim.append(rav[1])
+      else:
+        per_input_ragged_axis_dim.append(None)
 
     def when_wrapped_kernel(lengths_ref, *args, **kwargs):
-      b_idx = primitives.program_id(stacked_axis)
-      i_idx = (
-          primitives.program_id(ragged_axis_dim)
-          * val_at_ragged_dim
-      )
+      b_idx = primitives.program_id(mapped_dim_idx)
+
       b_len = lengths_ref[b_idx]
+      run_kernel = jnp.array(True)
+      for i, _ in enumerate(args):
+        ragged_axis_dim = per_input_ragged_axis_dim[i]
+        if ragged_axis_dim is None:
+          continue
+        arg_i_idx = (
+            primitives.program_id(ragged_axis_dim)
+            * block_shapes[i][ragged_axis_dim]
+        )
+        run_kernel = jnp.logical_and(run_kernel, arg_i_idx < b_len)
 
       # TODO(mvoz): Unimplemented primitive in pallas
       # b_len_mod = jnp.equal(jnp.mod(b_len, val_at_ragged_dim), 0)
       # checkify.check(b_len_mod, "b_len % val_at_ragged_dim != 0")
 
-      @pallas_utils.when(i_idx < b_len)
+      @pallas_utils.when(run_kernel)
       def f():
         # Important! This allows us to trace the inner kernel with the correct
         # grid to preserve user program_id semantics. Ex: program_id(0) will
@@ -922,20 +991,108 @@ def _pallas_call_batching_rule(
         with pallas_core.tracing_grid_env(grid_mapping.grid, ()):
           jax_core.eval_jaxpr(jaxpr, (), *args, **kwargs)
 
-      if debug_zero_fill_counterfactual:
-
-        @pallas_utils.when(i_idx >= b_len)
-        def g():
-          for arg_ref in args:
-            arg_ref[...] = jnp.zeros_like(arg_ref)
-
     kernel_avals = [lengths_aval] + [v.aval for v in jaxpr.invars]
     flat_kernel_avals, kernel_in_tree = tree_util.tree_flatten(
         list(kernel_avals)
     )
+
+    def _rewrite_index_jaxpr(enumerate_batched_block_mapping):
+      arg_pos, batched_block_mapping = enumerate_batched_block_mapping
+      indexer_avals = [
+          v.aval for v in batched_block_mapping.index_map_jaxpr.jaxpr.invars
+      ]
+      flat_indexer_avals, indexer_in_tree = tree_util.tree_flatten(
+          list(indexer_avals)
+      )
+
+      def index_rewrite_kernel(*indexer_args):
+        ragged_axis_dim = per_input_ragged_axis_dim[arg_pos]
+
+        # the problem here seems to be that we are rnning this for all inputs, per input, because they each have an indexer - which means
+        # that the indexer for output isnt getting written - before, it always was
+
+        lengths_ref = indexer_args[-1]
+        rest_indexer_args = indexer_args[:-1]
+        # Lengths are always the last argument of the indexer.
+        # lengths_ref = args[-1]
+        # Invariant: Stacked axis is enforced to be the mapped axis above.
+        b_idx = indexer_args[mapped_dim_idx]
+
+        nargs = list(rest_indexer_args)
+
+        if ragged_axis_dim is not None:
+          val_at_ragged_dim = batched_block_mapping.block_shape[ragged_axis_dim]
+
+          # The current index into the ragged dimension.
+          # Invariant: There is only one ragged dimension, enforced above.
+          i_idx = indexer_args[ragged_axis_dim]
+
+          # grid space -> element space
+          i_len = i_idx * val_at_ragged_dim
+
+          # The length of the current batch.
+          b_len = lengths_ref[b_idx]
+
+          # Have we reached the end of the current batch?
+          not_done = i_len < b_len
+
+          am_last_batch = b_idx == axis_size - 1
+          last_good_block = lax.div(b_len, val_at_ragged_dim) - 1
+
+          # The logic below can be thought of as:
+          #  if index_oob_ragged:
+          #    if not last_batch:
+          #      batch_idx += 1
+          #      ragged_idx = 0
+          #    else:
+          #      ragged_idx = last_good_block
+          #
+          # wherein we find the next good block by incrementing the batch index
+          # and setting the ragged index to 0 if we are not in the last batch.
+          # Otherwise, we set the ragged index to the last good block.
+          b_next = jnp.where(
+              not_done, b_idx, jnp.where(am_last_batch, b_idx, b_idx + 1)
+          )
+          i_next = jnp.where(
+              not_done, i_idx, jnp.where(am_last_batch, last_good_block, 0)
+          )
+          nargs[ragged_axis_dim] = i_next
+          nargs[mapped_dim_idx] = b_next
+
+        nargs = nargs + [lengths_ref]
+        return jax_core.eval_jaxpr(
+            batched_block_mapping.index_map_jaxpr.jaxpr,
+            batched_block_mapping.index_map_jaxpr.consts,
+            *nargs,
+        )
+
+      index_jaxpr, _ = _trace_kernel_to_jaxpr(
+          index_rewrite_kernel,
+          "index_rewrite_kernel",
+          batched_grid_mapping,
+          tuple(flat_indexer_avals),
+          indexer_in_tree,
+          tuple(() for _ in flat_indexer_avals),
+          interpret=interpret,
+          indexer=True,
+      )
+
+      batched_block_mapping = batched_block_mapping.replace(
+          index_map_jaxpr=pe.close_jaxpr(index_jaxpr)
+      )
+      return batched_block_mapping
+
     # Important! This allows us to trace the outer kernel with the correct grid
     # to enable accessing the batch program_id.
     with pallas_core.tracing_grid_env(batched_grid_mapping.grid, ()):
+      batched_block_mappings = map(
+          _rewrite_index_jaxpr, enumerate(batched_block_mappings)
+      )
+
+      batched_grid_mapping = batched_grid_mapping.replace(
+          block_mappings=tuple(batched_block_mappings),
+      )
+
       kernel_src_info: pallas_core.SrcInfoStr = "<Wrapped outer kernel>"
 
       jaxpr, consts = _trace_kernel_to_jaxpr(
@@ -950,7 +1107,15 @@ def _pallas_call_batching_rule(
       if consts:
         raise NotImplementedError("consts not supported in pallas_call")
 
-    assert ragged_axis_length is not None
+    # We need to rewrite the input_output_aliases here, the initial call
+    # to broadcast is done, and we have inseted a new input (lengths), so
+    # there's an off-by-one here now.
+    new_input_output_aliases = []
+    for k, v in input_output_aliases:
+      new_input_output_aliases.append((k + 1, v))
+    input_output_aliases = tuple(new_input_output_aliases)
+
+    # assert ragged_axis_length is not None
     args = (ragged_axis_length, *args)
   assert all(isinstance(aval, jax_core.ShapedArray) for aval in out_avals)
   batched_out_avals = tuple(
@@ -971,6 +1136,7 @@ def _pallas_call_batching_rule(
       compiler_params=compiler_params,
       cost_estimate=batched_cost_estimate,
       out_avals=batched_out_avals,
+      backend=backend,
   )
   return out, (0,) * len(out)
 
@@ -993,6 +1159,85 @@ def checkify_pallas_kernel_body_jaxpr(
         body_jaxpr, enabled_errors, err_tree, *flat_err_and_in_vals)
   return checked_jaxpr, out_tree, error_effects
 
+def pallas_call_checkify_oob_grid(error: checkify.Error,
+                                  enabled_errors,
+                                  args: jax_core.Value,
+                                  grid_mapping: GridMapping,
+                                  input_output_aliases) -> checkify.Error:
+  if checkify.OOBError not in enabled_errors:
+    return error
+  dynamic_grid_args, args = split_list(
+      args, [grid_mapping.num_dynamic_grid_bounds]
+  )
+  output_args = _initialize_output_vals(grid_mapping.block_mappings_output,
+                                    args, input_output_aliases)
+  scalars, input_args, _ = split_list(
+      args, [grid_mapping.num_index_operands,
+             grid_mapping.num_inputs],
+  )
+  dynamic_grid_args_iter = iter(dynamic_grid_args)
+  grid = tuple(
+      a if a is not pallas_core.dynamic_grid_dim
+      else next(dynamic_grid_args_iter)
+      for a in grid_mapping.grid
+  )
+  grid_start_indices = (jnp.int32(0),) * len(grid)
+  if grid:
+    num_iterations = reduce(jnp.multiply, grid)  # type: ignore[arg-type]
+  else:
+    # Base case is always one iteration when grid is ()
+    num_iterations = 1
+
+  is_indexing_dim = [
+      tuple(b is pallas_core.mapped for b in bm.block_shape)
+      for bm in grid_mapping.block_mappings
+  ]
+  block_shapes = [
+      None if iid is None
+      else tuple(1 if i else b for i, b in zip(iid, bm.block_shape))
+      for iid, bm in zip(is_indexing_dim, grid_mapping.block_mappings)
+  ]
+  # The scan carry: (i, loop_idx, *consts, *ins, *outs, *scratch)
+  # i:int32 is the interation index
+  # loop_idx: tuple[int32] are the program ids for each grid axis
+  def cond(carry):
+    i, *_ = carry
+    return i < num_iterations
+  def body(carry):
+    i, loop_idx = carry
+    if grid_mapping.local_grid_env is not None:
+      local_grid_env = grid_mapping.local_grid_env(loop_idx, grid)
+    else:
+      local_grid_env = tuple(
+          pallas_core.GridAxis(idx, b)
+          for dim, (idx, b) in enumerate(zip(loop_idx, grid))
+          if dim not in grid_mapping.vmapped_dims
+      )
+    with pallas_core.grid_env(local_grid_env):
+      start_indices = [
+          None if bm is None else bm.compute_start_indices_interpret(loop_idx, *scalars)
+          for bm in grid_mapping.block_mappings]
+    # We perform a dynamic slice on the i/o blocks, which will be checked by
+    # checkify for OOB accesses.
+    map(_maybe_dynamic_slice, start_indices, block_shapes,
+        [*input_args, *output_args], is_indexing_dim)
+    return (i + 1, _get_next_indices(grid, loop_idx))
+  def f(_):
+    lax.while_loop(
+        cond, body, (jnp.int32(0), grid_start_indices)
+    )
+  flat_args, jaxpr_in_tree = jax.tree_util.tree_flatten((jnp.int32(0),))
+  wrapped_loop, _ = api_util.flatten_fun_nokwargs(
+      lu.wrap_init(f), jaxpr_in_tree)
+  with pallas_core.tracing_grid_env(grid_mapping.grid, ()):
+    avals_in = map(jax_core.get_aval, flat_args)
+    traced_loop, _, consts, () = pe.trace_to_jaxpr_dynamic(
+        wrapped_loop, list(avals_in))
+    traced_loop = jax_core.ClosedJaxpr(traced_loop, consts)
+  out_error, _ = checkify.checkify_jaxpr(
+      traced_loop, checkify.index_checks, error, flat_args)
+  return out_error
+
 def pallas_call_checkify_rule(error: checkify.Error,
                               enabled_errors,
                               *args: jax_core.Value,
@@ -1002,6 +1247,10 @@ def pallas_call_checkify_rule(error: checkify.Error,
                               grid_mapping: GridMapping,
                               out_avals: tuple[jax_core.AbstractValue, ...],
                               **kwargs):
+  # Check for OOB accesses in the grid.
+  error = pallas_call_checkify_oob_grid(error, enabled_errors,
+                                        args, grid_mapping,
+                                        input_output_aliases)
   # We implement the checkify rule in 4 steps:
   # 1) First, trace the kernel body to get the expected error shapes.
   # 2) Checkify the kernel body to obtain a jaxpr with errors as inputs
@@ -1143,18 +1392,6 @@ def pallas_call_checkify_rule(error: checkify.Error,
 checkify.error_checks[pallas_call_p] = pallas_call_checkify_rule
 
 
-# All of those shenanigans are because we can't make TransformedRef a PyTree,
-# because they should appear as atomic JAX values to the users.
-@lu.transformation
-def wrap_with_transforms(transforms, *args):
-  new_args = tuple(
-      state_types.TransformedRef(a, t) if t else a
-      for a, t in zip(args, transforms)
-  )
-  res = yield new_args, {}
-  yield res
-
-
 @weakref_lru_cache
 def _trace_kernel_to_jaxpr(
     fun: Callable,
@@ -1164,13 +1401,16 @@ def _trace_kernel_to_jaxpr(
     kernel_in_tree: tree_util.PyTreeDef,
     kernel_in_transforms: tuple[tuple[pallas_core.Transform, ...], ...],
     interpret: bool,
+    indexer: bool = False,
 ) -> tuple[jax_core.ClosedJaxpr, tuple[jax.Array, ...]]:
   if interpret:
     kernel_avals = tuple(map(_logical_aval_to_interpret_mode_aval,
                              kernel_avals))
   wrapped_kernel_fun, out_tree_thunk = api_util.flatten_fun_nokwargs(
       lu.wrap_init(fun), kernel_in_tree)
-  wrapped_kernel_fun = wrap_with_transforms(wrapped_kernel_fun, kernel_in_transforms)
+  wrapped_kernel_fun = primitives.wrap_with_transforms(
+      wrapped_kernel_fun, kernel_in_transforms
+  )
   debug = pe.debug_info(fun, kernel_in_tree, out_tree_thunk, False, "pallas_call")
   with grid_mapping.trace_env():
     jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(wrapped_kernel_fun,
@@ -1185,7 +1425,7 @@ def _trace_kernel_to_jaxpr(
             "You should pass them as inputs")
 
   kernel_out_tree = out_tree_thunk()
-  if kernel_out_tree != tree_util.tree_structure(None):
+  if not indexer and kernel_out_tree != tree_util.tree_structure(None):
     raise ValueError(
         f"The kernel function in the pallas_call {name_and_src_info} "
         f"should return None. It returns a PyTree: {kernel_out_tree}")
@@ -1210,9 +1450,15 @@ def _unsupported_lowering_error(platform: str) -> Exception:
       " https://jax.readthedocs.io/en/latest/installation.html."
   )
 
+_Backend = Literal["mosaic_tpu", "triton", "mosaic_gpu"]
+
 
 def _pallas_call_lowering(
-    ctx: mlir.LoweringRuleContext, *in_nodes, interpret: bool, **params
+    ctx: mlir.LoweringRuleContext,
+    *in_nodes,
+    interpret: bool,
+    backend: _Backend | None,
+    **params,
 ):
   if params['jaxpr'].constvars:
     raise ValueError('Cannot lower a pallas_call with constants.')
@@ -1229,6 +1475,8 @@ def _pallas_call_lowering(
   def tpu_lowering(ctx: mlir.LoweringRuleContext,
                    *in_nodes: mlir.ir.Value | Sequence[mlir.ir.Value],
                    **params):
+    if backend and backend != "mosaic_tpu":
+      raise ValueError("Only mosaic backend supported for TPU")
     if mosaic_tpu_backend is None:
       raise _unsupported_lowering_error("tpu")
     return mosaic_tpu_backend.pallas_call_tpu_lowering_rule(
@@ -1239,12 +1487,21 @@ def _pallas_call_lowering(
                    *in_nodes: mlir.ir.Value | Sequence[mlir.ir.Value],
                    **params):
     try:
-      if _PALLAS_USE_MOSAIC_GPU.value:
-        from jax._src.pallas.mosaic_gpu import pallas_call_registration
-      else:
-        from jax._src.pallas.triton import pallas_call_registration  # type: ignore
-    except ImportError:
+      match backend:
+        case "mosaic_gpu":
+          from jax._src.pallas.mosaic_gpu import pallas_call_registration
+        case "triton":
+          from jax._src.pallas.triton import pallas_call_registration  # type: ignore
+        case None:
+          if _PALLAS_USE_MOSAIC_GPU.value:
+            from jax._src.pallas.mosaic_gpu import pallas_call_registration
+          else:
+            from jax._src.pallas.triton import pallas_call_registration  # type: ignore
+        case _:
+          raise ValueError(f"Unsupported backend: {backend}")
+    except ImportError as e:
       raise _unsupported_lowering_error("gpu")
+
     return pallas_call_registration.pallas_call_lowering(
         ctx, *in_nodes, **params
     )
@@ -1313,6 +1570,7 @@ def _pallas_call_state_discharge_rule(
     compiler_params: Any,
     cost_estimate: CostEstimate | None,
     out_avals: tuple[jax_core.AbstractValue, ...],
+    backend: _Backend | None = None
 ):
   del avals_out
   assert all(isinstance(v.aval, state.AbstractRef) for v in jaxpr.constvars)
@@ -1414,6 +1672,7 @@ def _pallas_call_state_discharge_rule(
       compiler_params=compiler_params,
       cost_estimate=cost_estimate,
       out_avals=new_out_avals,
+      backend=backend,
   )
   refs_out, rest = split_list(out_flat, [num_refs])
   updated_vals_in = refs_out + [None] * len(rest_in_avals)
@@ -1435,6 +1694,7 @@ def pallas_call(
     name: str | None = None,
     compiler_params: dict[str, Any] | pallas_core.CompilerParams | None = None,
     cost_estimate: CostEstimate | None = None,
+    backend: _Backend | None = None,
 ) -> Callable[..., Any]:
   """Invokes a Pallas kernel on some inputs.
 
@@ -1484,6 +1744,8 @@ def pallas_call(
       platform is either 'mosaic' or 'triton'. It is also possible
       to pass in `jax.experimental.pallas.tpu.TPUCompilerParams` for TPUs and
       `jax.experimental.pallas.gpu.TritonCompilerParams` for Triton/GPUs.
+    backend: Optional string literal one of  "mosaic_tpu", "triton" or "mosaic_gpu"
+      determining the backend to be used. None means let pallas decide.
 
 
   Returns:
@@ -1590,6 +1852,18 @@ def pallas_call(
             f"[0, {len(flat_out_avals)})")
       in_aval = flat_in_avals[i_idx]
       out_aval = flat_out_avals[o_idx]
+      if isinstance(in_aval, jax_core.DShapedArray):
+        new_shape = []
+        for d in in_aval.shape:
+          if isinstance(d, int):
+            new_shape.append(d)
+          else:
+            new_shape.append(d.dtype.bound)
+
+        in_aval = jax_core.ShapedArray(
+            tuple(new_shape), in_aval.dtype, in_aval.weak_type
+        )
+
       if in_aval.shape != out_aval.shape or in_aval.dtype != out_aval.dtype:
         raise ValueError(
             f"input_output_aliases contains the mapping '{i_idx}:{o_idx}' "
@@ -1614,6 +1888,7 @@ def pallas_call(
           input_output_aliases=tuple(input_output_aliases.items()),
           compiler_params=compiler_params,
           cost_estimate=cost_estimate,
+          backend=backend,
       )
     out = tree_util.tree_unflatten(out_tree, out_flat)
     return out
