@@ -202,6 +202,9 @@ _plugin_callback_lock = threading.Lock()
 # for unimplemented features. Wrong outputs are not acceptable.
 _nonexperimental_plugins: set[str] = {'cuda', 'rocm'}
 
+# The set of known experimental plugins that have registrations in JAX codebase.
+_experimental_plugins: set[str] = {"METAL"}
+
 def register_backend_factory(name: str, factory: BackendFactory, *,
                              priority: int = 0,
                              fail_quietly: bool = True,
@@ -774,12 +777,20 @@ for _platform, _alias in _platform_aliases.items():
   _alias_to_platforms.setdefault(_alias, []).append(_platform)
 
 
+def known_platforms() -> set[str]:
+  platforms = set()
+  platforms |= set(_nonexperimental_plugins)
+  platforms |= set(_experimental_plugins)
+  platforms |= set(_backend_factories.keys())
+  platforms |= set(_platform_aliases.values())
+  return platforms
+
+
 def is_known_platform(platform: str) -> bool:
   # A platform is valid if there is a registered factory for it. It does not
   # matter if we were unable to initialize that platform; we only care that
   # we've heard of it and it isn't, e.g., a typo.
-  return (platform in _backend_factories.keys() or
-          platform in _platform_aliases.keys())
+  return platform in known_platforms()
 
 
 def canonicalize_platform(platform: str) -> str:
