@@ -70,16 +70,40 @@ class WGSplatFragLayout:
     """
     return all(dim1 == dim2 or dim1 == 1 for dim1, dim2 in zip(self.shape[::-1], shape[::-1]))
 
+  def thread_idxs(self, shape):
+    assert shape == self.shape
+    raise NotImplementedError
+
 
 @dataclasses.dataclass(frozen=True)
 class WGMMAFragLayout:
   """[m, n] matrix, where m % 64 == 0 == n % 8."""
 
+  def thread_idxs(self, shape):
+    index = ir.IndexType.get()
+    assert shape[0] % 64 == 0 and shape[1] % 8 == 0
+    tid = arith.index_cast(ir.IndexType.get(), mgpu.thread_idx())
+    tid_wg = arith.remui(tid, c(WARPGROUP_SIZE, index))
+    warp_idx = arith.divui(tid_wg, c(32, index))
+    tid_warp = arith.remui(tid_wg, c(32, index))
+    col_base = arith.muli(arith.remui(tid_warp, c(4, index)), c(2, index))
+    row_base = arith.addi(
+        arith.divui(tid_warp, c(4, index)), arith.muli(warp_idx, c(16, index))
+    )
+    for row_group in range(0, shape[0], 64):
+      for col_group in range(0, shape[1], 8):
+        for row_subgroup in range(0, 16, 8):
+          row = arith.addi(row_base, c(row_group + row_subgroup, index))
+          yield row, arith.addi(col_base, c(col_group, index))
+
 
 @dataclasses.dataclass(frozen=True)
 class WGMMARowFragLayout:
   """[m] matrix, where m % 64 == 0."""
 
+  def thread_idxs(self, shape):
+    raise NotImplementedError
+
 
 @dataclasses.dataclass(frozen=True)
 class WGStridedFragLayout:
@@ -110,9 +134,10 @@ class WGStridedFragLayout:
         shape=tuple(memref_type.shape), vec_size=min(8 // bw, max_vec_size)
     )
 
-  def thread_vec_idxs(self):
+  def thread_idxs(self, shape):
+    assert shape == self.shape
     index = ir.IndexType.get()
-    for v in self.linear_thread_vec_idxs():
+    for v in self.linear_thread_idxs():
       res = []
       for dim in reversed(self.shape):
         dim = c(dim, index)
@@ -121,7 +146,7 @@ class WGStridedFragLayout:
       res.reverse()
       yield res
 
-  def linear_thread_vec_idxs(self):
+  def linear_thread_idxs(self):
     """The indexes to be used for vector load/store WGStridedFragLayout.
 
     Yields:
@@ -214,10 +239,17 @@ class FragmentedArray:
       raise TypeError(ref.type)
 
     ref_ty = ir.MemRefType(ref.type)
-    ref_1d = mgpu.memref_fold(ref, 0, len(ref_ty.shape))
+    shape = tuple(ref_ty.shape)
     layout = WGStridedFragLayout.from_memref_type(ref_ty)
     vec_ty = ir.VectorType.get((layout.vec_size,), ref_ty.element_type)
-    vecs = [vector.load(vec_ty, ref_1d, [vec_idx]) for vec_idx in layout.linear_thread_vec_idxs()]
+    try:
+      # Flattening the reference potentially produces simpler PTX but
+      # if the ref is not already 1D and has strided dimensions
+      # flattening won't work.
+      ref_ = mgpu.memref_fold(ref, 0, len(ref_ty.shape))
+      vecs = [vector.load(vec_ty, ref_, [vec_idx]) for vec_idx in layout.linear_thread_idxs()]
+    except NotImplementedError:
+      vecs = [vector.load(vec_ty, ref, vec_idx) for vec_idx in layout.thread_idxs(shape)]
     return cls(_registers=np.array(vecs), _layout=layout, _is_signed=is_signed)
 
   @classmethod
@@ -226,15 +258,20 @@ class FragmentedArray:
     match layout:
       case WGMMARowFragLayout():
         if len(shape) != 1:
-          raise ValueError
+          raise ValueError("WGMMARowFragLayout requires a 1D shape")
         if shape[0] % 64:
-          raise ValueError
+          raise ValueError(
+              "WGMMARowFragLayout requires shape[0] to be a multiple of 64"
+          )
         reg_shape = (shape[0] // 64, 2)
       case WGMMAFragLayout():
         if len(shape) != 2:
-          raise ValueError
+          raise ValueError("WGMMAFragLayout requires a 2D shape")
         if shape[0] % 64 or shape[1] % 8:
-          raise ValueError
+          raise ValueError(
+              "WGMMAFragLayout requires shape[0] to be a multiple of 64, and"
+              " shape[1] to be a multiple of 8"
+          )
         reg_shape = (shape[0] // 64, shape[1] // 8, 2, 1)
         value = vector.splat(ir.VectorType.get((2,), value.type), value)
       case WGStridedFragLayout(vec_size=vec_size):
@@ -277,6 +314,22 @@ class FragmentedArray:
       case WGMMARowFragLayout() | WGSplatFragLayout():
         return reg_ty
 
+  def to_layout(self, new_layout: FragmentedLayout):
+    """Converts the fragmented array to the given layout.
+
+    At the moment, only conversions from ``WGSplatFragLayout`` are supported.
+    """
+    if self.layout == new_layout:
+      return self
+    if not isinstance(self.layout, WGSplatFragLayout):
+      raise NotImplementedError(
+          f"Cannot convert from {self.layout} to {new_layout}"
+      )
+    [reg] = self.registers.flat
+    return type(self).splat(
+        reg, self.shape, new_layout, is_signed=self.is_signed
+    )
+
   def _pointwise(self, op, *other, output_is_signed: bool | None = None):
     is_signed = (
         output_is_signed if output_is_signed is not None else self.is_signed
@@ -659,6 +712,7 @@ class FragmentedArray:
         _registers=new_registers, _layout=self.layout, _is_signed=is_signed
     )
 
+  # NOTE: scratch can be reused immediately once this function returns.
   def reduce_sum(self, scratch) -> ir.Value:
     if ir.FloatType.isinstance(self.mlir_dtype):
       op = arith.addf
@@ -696,7 +750,9 @@ class FragmentedArray:
       )
       memref.store(scratch_sum, scratch, [zero_index])
     utils.warpgroup_barrier()
-    return memref.load(scratch, [zero_index])
+    result = memref.load(scratch, [zero_index])
+    utils.warpgroup_barrier()  # Make sure everyone is done using scratch.
+    return result
 
   def reduce(self, op, axis):
     if self.layout != WGMMA_LAYOUT:
@@ -794,12 +850,11 @@ class FragmentedArray:
 
   def foreach(self, fn: Callable[[ir.Value, tuple[ir.Value, ...]], None]):
     """Call a function for each value and index."""
-    if not isinstance(self.layout, WGStridedFragLayout):
-      raise NotImplementedError(self.layout)
     index = ir.IndexType.get()
-    for idx, reg in zip(self.layout.thread_vec_idxs(), self.registers.flat):
+    for idx, reg in zip(self.layout.thread_idxs(self.shape), self.registers.flat, strict=True):
       assert len(idx) == len(self.shape), (idx, self.shape)
-      for i in range(self.layout.vec_size):
+      [elems] = ir.VectorType(reg.type).shape
+      for i in range(elems):
         i = c(i, index)
         fn(vector.extractelement(reg, position=i), (*idx[:-1], arith.addi(idx[-1], i)))
 
@@ -835,12 +890,22 @@ class FragmentedArray:
 
   def _store_untiled_wg_strided(self, ref: ir.Value):
     ref_ty = ir.MemRefType(ref.type)
+    try:
+      # Flattening the reference potentially produces simpler PTX but
+      # if the ref is not already 1D and has strided dimensions
+      # flattening won't work. We use a different variable for ref in
+      # case `NotImplementedError` is thrown by
+      # .linear_thread_idxs().
+      ref_ = mgpu.memref_fold(ref, 0, len(ref_ty.shape))
+      idxs = ([i] for i in self.layout.linear_thread_idxs())
+    except NotImplementedError:
+      ref_ = ref
+      idxs = self.layout.thread_idxs()
     ref_shape = tuple(ref_ty.shape)
     if ref_shape != self.shape:
       raise ValueError((ref_shape, self.shape))
-    smem_1d = mgpu.memref_fold(ref, 0, len(ref_ty.shape))
-    for idx, reg in zip(self.layout.linear_thread_vec_idxs(), self.registers.flat):
-      vector.store(reg, smem_1d, [idx])
+    for idx, reg in zip(idxs, self.registers.flat):
+      vector.store(reg, ref_, idx)
 
   def _store_untiled_wgmma(self, ref: ir.Value):
     """Stores accumulator to a 2D memref. Not optimized at the moment."""
